/*
 *  -*- Mode: ANSI C -*-
 *  $Id: pbm.c,v 1.4 1996/09/17 16:10:28 fernande Exp $
 *  $Source: /sgi.acct/sweldens/cvs/liftpack/Files/pbm.c,v $
 *  Author: Gabriel Fernandez
 *
 *  Routines to manipulate 'p?m' format pictures.
 *  LoadPBM ( filename, image ) - loads a PBM, PGM, or PPM file.
 *  WritePBM ( fp, pic, w, h, style, raw, cmt, comment );
 */
/* do not edit anything above this line */

/* System header files */
#include <string.h>
/* FLWT header files */
#include <flwterr.h>
#include <file.h>
#include <imgmem.h>
#include <memchk.h>
#include <pbm.h>
#include <util.h>

/*
 * Comments on error handling:
 *    A truncated file is not considered a Major Error. The file is
 *    loaded, the rest of the pic is filled with 0's.
 *
 *    A file with garbage characters in it is an unloadable file. All
 *    allocated stuff is tossed, and LoadPBM returns non-zero.
 *
 *    Not being able to malloc is a Fatal Error. The program is aborted.
 */

#define TRUNCSTR "File appears to be truncated.\n"

/* Static variables */
static int garbage;
static long numgot, filesize;

/* Static functions declarations */
static int loadpbm  (FILE *, Image *, int);
static int loadpgm  (FILE *, Image *, int, int);
static int loadppm  (FILE *, Image *, int, int);
static int getint   (FILE *, Image *);
static int getbit   (FILE *, Image *);
static int getshort (FILE *);

/* code */

/*
 * LoadPBM function: loads a PNM file. Returns '1' on success.
 */
int
LoadPBM ( const char *filename, Image *image )
{
    FILE  *fp;
    int    c, c1;
    int    maxv, rv;

    garbage = maxv = rv = 0;

    image->band[RED]   = (Matrix)NULL;
    image->band[GREEN] = (Matrix)NULL;
    image->band[BLUE]  = (Matrix)NULL;
    image->comment     = (char *)NULL;
    image->levels      = (int  *)NULL;


    /* Open the file */
    fp = OpenFile (filename, "r");

    /* Compute file length */
    fseek (fp, 0L, 2);
    filesize = ftell(fp);
    fseek (fp, 0L, 0);


    /* Read the first two bytes of the file to determine which format
       this file is.  "P1" = ascii bitmap, "P2" = ascii greymap,
       "P3" = ascii pixmap, "P4" = raw bitmap, "P5" = raw greymap,
       "P6" = raw pixmap */

    c = getc(fp);  c1 = getc(fp);
    if (c!='P' || c1<'1' || c1>'6') {
        Error ("LoadPBM", CUSTOM, RETURN, "Invalid magic number.\n");
        return 0;
    }

    /* Read in header information */
    image->width  = getint(fp, image);
    image->height = getint(fp, image);

    /* if we're not reading a bitmap, read the 'max value' */
    if ( !(c1=='1' || c1=='4') ) {
        maxv = getint(fp, image);
        if ( maxv<1 )
             garbage = 1;    /* to avoid 'div by zero' probs */
    }

    if (garbage) {
        fclose(fp);
        if (image->comment)
            free(image->comment);
        image->comment = (char *)NULL;
        Error ("LoadPBM", CUSTOM, RETURN, "Garbage characters in header.\n");
        return 0;
    }


    if ( c1=='1' || c1=='2' || c1=='3' )
        image->frmType = F_PBMASCII;
    else
        image->frmType = F_PBMRAW;

    /* Note:  pic, band, bpp and colorPlanes fields of
       Image struct are filled in in the format-specific loaders */

    /* Call the appropriate subroutine to handle format-specific stuff */
    if ( c1=='1' || c1=='4' )
        rv = loadpbm(fp, image, c1=='4' ? 1 : 0);
    else if ( c1=='2' || c1=='5' )
        rv = loadpgm(fp, image, c1=='5' ? 1 : 0, maxv);
    else if ( c1=='3' || c1=='6' )
        rv = loadppm(fp, image, c1=='6' ? 1 : 0, maxv);

    fclose(fp);

    if (!rv) {
        if (image->band[RED])
            MEM_Free2D(image->band[RED]);
        if (image->band[GREEN])
            MEM_Free2D(image->band[GREEN]);
        if (image->band[BLUE])
            MEM_Free2D(image->band[BLUE]);
        if (image->comment)
            free(image->comment);
        image->band[RED]   = (Matrix)NULL;
        image->band[GREEN] = (Matrix)NULL;
        image->band[BLUE]  = (Matrix)NULL;
        image->comment     = (char *)NULL;
    }

    return rv;
}  


/*
 * loadpbm function: reads the PBM file pointed by fp.
 */
static int loadpbm ( FILE *fp, Image *image, int raw )
{
    Vector pix;
    int    i, j, bit, w, h, rv;

    /******************************/
    /* Initialize image structure */
    /******************************/
    rv = (int)IMG_InitMem (image, (long)image->width, (long)image->height, 1, 1);
    if ( !rv ) {
        Error ("loadpbm", MEMORY_IMAGE, ABORT);
        /* NOTREACHED */
    }
    image->colType = 2;   /* B/W stipple */
    w = image->width;
    h = image->height;
    sprintf(image->fullInfo, "PBM, %s format.  (%ld bytes)", 
	    (raw) ? "raw" : "ascii", filesize);
    sprintf(image->shrtInfo, "%dx%d PBM.", w, h);


    if ( !raw ) {
        numgot = 0;
        for ( i=0, pix=image->band[GREY][0] ; i<h ; i++ ) {
            for ( j=0 ; j<w ; j++, pix++ )
                *pix = (Flt)getbit(fp, image);
        }

        if ( numgot != (long)(w*h) )
            Error ("loadpbm", CUSTOM, RETURN, TRUNCSTR);
        if (garbage) {
            Error ("loadpbm", CUSTOM, RETURN,
                   "Garbage characters in image data.");
            return 0;
        }
    } else {   /* read raw bits */
        int trunc = 0, k = 0;

        for ( i=0, pix=image->band[GREY][0] ; i<h ; i++ ) {
            for ( j=0, bit=0 ; j<w ; j++, pix++, bit++ ) {
	        bit &= 7;
	        if ( !bit ) {
	            k = getc(fp);
	            if ( k==EOF ) { trunc = 1; k = 0; }
	        }
	        *pix = (k&0x80) ? (Flt)1 : (Flt)0;
	        k = k << 1;
            }
        }

        if (trunc)
            Error ("loadpbm", CUSTOM, RETURN, TRUNCSTR);
    }

    return 1;
}


/*
 * loadpgm function: reads the PGM file pointed by fp.
 */
static int loadpgm ( FILE *fp, Image *image, int raw, int maxv )
{
    Vector  pixFlt;
    byte   *pic8;
    int     i, j, bitshift, w, h, holdmaxv, rv;

    /******************************/
    /* Initialize image structure */
    /******************************/
    rv = (int)IMG_InitMem (image, (long)image->width, (long)image->height, 1, 8);
    if ( !rv ) {
        Error ("loadpgm", MEMORY_IMAGE, ABORT);
        /* NOTREACHED */
    }
    image->colType = 1;   /* Greyscale */
    w = image->width;
    h = image->height;
    sprintf(image->fullInfo, "PGM, %s format.  (%ld bytes)", 
	    (raw) ? "raw" : "ascii", filesize);
    sprintf(image->shrtInfo, "%dx%d PGM.", image->width, image->height);


    /* if maxv>255, keep dropping bits until it's reasonable */
    holdmaxv = maxv;
    bitshift = 0;
    while ( maxv>255 ) { maxv = maxv>>1;  bitshift++; }


    numgot = 0;

    if (!raw) {
        for ( i=0, pixFlt=image->band[GREY][0] ; i<h ; i++ ) {
            for ( j=0; j<w ; j++, pixFlt++ )
                *pixFlt = (Flt) (getint(fp, image) >> bitshift);
        }
    } else { /* raw */
        if ( holdmaxv>255 ) {
            for ( i=0, pixFlt=image->band[GREY][0] ; i<h ; i++ ) {
                for ( j=0 ; j<w ; j++, pixFlt++ )
                    *pixFlt = (Flt) (getshort(fp) >> bitshift);
            }
        } else {    /* read raw data */
            pic8 = (byte *) calloc((size_t)(w*h), (size_t)1);
            if ( !pic8 )
                Error ("loadpgm", CALLOC_FAIL, ABORT);
                /* NOTREACHED */
            numgot = (long)fread(pic8, (size_t)1, (size_t)(w*h), fp);
            for ( i=0, pixFlt=image->band[GREY][0] ; i<w*h ; i++, pixFlt++ )
                *pixFlt = (Flt)pic8[i];
            free (pic8);
        }
    }

    if ( numgot != (long)(w*h) ) {
        Error ("loadpgm", CUSTOM, RETURN, TRUNCSTR);   /* warning only */
    }

    if (garbage) {
        Error ("loadpgm", CUSTOM, RETURN, "Garbage characters in image data.");
        return 0;
    }

    return 1;
}


/*
 * loadppm function: reads the PPM file pointed by fp.
 */
static int loadppm ( FILE *fp, Image *image, int raw, int maxv )
{
    byte *pix, *pic24;
    byte scale[256];
    int   i, j, bitshift, w, h, holdmaxv, rv;

    w = image->width;
    h = image->height;

    /* allocate 24-bit image */
    pic24 = (byte *) calloc((size_t)(w*h*3), (size_t) 1);
    if ( !pic24 )
        Error ("loadppm", CALLOC_FAIL, ABORT);
        /* NOTREACHED */

    /******************************/
    /* Initialize image structure */
    /******************************/
    rv = (int)IMG_InitMem (image, (long)image->width, (long)image->height, 3, 24);
    if ( !rv ) {
        Error ("loadppm", MEMORY_IMAGE, ABORT);
        /* NOTREACHED */
    }
    image->colType = 0;   /* Full Color */
    sprintf(image->fullInfo, "PPM, %s format.  (%ld bytes)", 
	    (raw) ? "raw" : "ascii", filesize);
    sprintf(image->shrtInfo, "%dx%d PPM.", w, h);


    /* If maxv>255, keep dropping bits until it's reasonable */
    holdmaxv = maxv;
    bitshift = 0;
    while ( maxv>255 ) { maxv = maxv>>1;  bitshift++; }


    numgot = 0;

    if (!raw) {
        for ( i=0, pix=pic24; i<h ; i++ ) {
            for ( j=0; j<w*3; j++, pix++ )
	        *pix = (byte) (getint(fp, image) >> bitshift);
        }
    } else { /* raw */
        if ( holdmaxv>255 ) {
            for ( i=0, pix=pic24 ; i<h ; i++ ) {
	        for ( j=0; j<w*3; j++, pix++ )
	            *pix = (byte) (getshort(fp) >> bitshift);
            }
        } else {   /* read data */
            numgot = (long)fread(pic24, (size_t)1, (size_t)(w*h*3), fp);
        }
    }
  
    if ( numgot != (long)(w*h*3) ) {
        Error ("loadppm", CUSTOM, RETURN, TRUNCSTR);
    }

    if (garbage) {
        Error ("loadppm", CUSTOM, RETURN, "Garbage characters in image data.");
        return 0;
    }


    /* Scale all RGB values up to range 0-255 */
    if ( maxv<255 ) { 
        for ( i=0 ; i<=maxv ; i++)
            scale[i] = (byte)( (i * 255) / maxv );

        for ( i=0, pix=pic24 ; i<h ; i++ ) {
            for ( j=0 ; j<w*3 ; j++, pix++ )
                *pix = scale[*pix];
        }
    }

    /* Copy byte values to image bands */
    for ( i=0, pix=pic24 ; i<h ; i++ ) {
        for ( j=0 ; j<w ; j++, pix+=3 ) {
            image->band[RED][i][j]   = (Flt)(pix[0]);
            image->band[GREEN][i][j] = (Flt)(pix[1]);
            image->band[BLUE][i][j]  = (Flt)(pix[2]);
        }
    }
    free ( (byte *)pic24 );

    return 1;
}


/*
 * getint function: gets and integer value from the actual position
 *                  pointed by fp.
 */
static int getint ( FILE *fp, Image *image )
{
    int c, i, firstchar;

    /* Note:  if it sees a '#' character, all characters from there to end of
       line are appended to the comment string */

    /* Skip forward to start of next number */
    c = getc(fp);
    while (1) {
        /* Eat comments */
        if ( c=='#' ) {   /* if we're at a comment, read to end of line */
            char cmt[256], *sp, *tmpptr;

            sp = cmt; firstchar = 1;
            while (1) {
	        c = getc(fp);
	        if ( firstchar && c == ' ' )
                    firstchar = 0;  /* loop off 1 sp after # */
	        else {
	            if ( c == '\n' || c == EOF )
                        break;
	            if ( (sp-cmt)<250 )
                        *sp++ = (char)c;
	        }
            }
            *sp++ = '\n';
            *sp   = '\0';

            if ( strlen(cmt) > 0 ) {    /* add to image->comment */
	        if ( !image->comment ) {
	            image->comment = (char *) malloc(strlen(cmt) + 1);
	            if ( !image->comment )
                        Error ("getint", MALLOC_FAIL, ABORT);
                        /* NOTREACHED */
	            image->comment[0] = '\0';
	        } else {
	            tmpptr = (char *) realloc(image->comment, 
		             strlen(image->comment) + strlen(cmt) + 1); 
	            if (!tmpptr)
                        Error ("getint", REALLOC_FAIL, ABORT);
                        /* NOTREACHED */
	            image->comment = tmpptr;
	        }
	        strcat(image->comment, cmt);
            }
        }

        if ( c==EOF )
            return 0;
        if ( c>='0' && c<='9' )
            break;   /* we've found what we were looking for */

        /* see if we are getting garbage (non-whitespace) */
        if ( c!=' ' && c!='\t' && c!='\r' && c!='\n' && c!=',' )
            garbage=1;

        c = getc(fp);
    }


    /* we're at the start of a number, continue until we hit a non-number */
    i = 0;
    while (1) {
        i = (i*10) + (c - '0');
        c = getc(fp);
        if ( c==EOF )
            return i;
        if ( c<'0' || c>'9' )
            break;
    }

    numgot++;
    return i;
}


/*
 * getshort function: it is used in RAW mode to read 16-bit values.
 */
static int getshort ( FILE *fp )
{
    int c1, c2;

    c1 = getc(fp);
    if ( c1 == EOF )
        return 0;
    c2 = getc(fp);
    if ( c2 == EOF )
        return 0;

    numgot++;

    return (c2 << 8) | c1;
}


/*
 * getbit function: reads the next bit in stream pointed by fp.
 */
static int getbit ( FILE *fp, Image *image )
{
    int c;

    /* skip forward to start of next number */
    c = getc(fp);
    while (1) {
        /* eat comments */
        if ( c=='#' ) {   /* if we're at a comment, read to end of line */
            char cmt[256], *sp, *tmpptr;

            sp = cmt;
            while (1) {
	        c = getc(fp);
	        if ( c == '\n' || c == EOF )
                    break;

	        if ( (sp-cmt)<250 )
                    *sp++ = (char)c;
            }
            *sp++ = '\n';
            *sp   = '\0';

            if ( strlen(cmt) > 0 ) {    /* add to image->comment */
	        if ( !image->comment ) {
	            image->comment = (char *) malloc(strlen(cmt) + 1);
	            if ( !image->comment )
                        Error ("getbit", MALLOC_FAIL, ABORT);
                        /* NOTREACHED */
	            image->comment[0] = '\0';
	        } else {
	            tmpptr = (char *) realloc(image->comment, 
		             strlen(image->comment) + strlen(cmt) + 1); 
	            if (!tmpptr)
                        Error ("getbit", REALLOC_FAIL, ABORT);
                        /* NOTREACHED */
	            image->comment = tmpptr;
	        }
	        strcat(image->comment, cmt);
            }
        }
        if ( c==EOF )
            return 0;
        if ( c=='0' || c=='1' )
            break;   /* we've found what we were looking for */

        /* see if we are getting garbage (non-whitespace) */
        if ( c!=' ' && c!='\t' && c!='\r' && c!='\n' && c!=',' )
            garbage=1;

        c = getc(fp);
    }

    numgot++;
    return (c-'0');
}


/*
 * WritePBM function: writes a PBM/PGM/PPM file to the already open stream.
 *                    If (raw), writes as RAW bytes, otherwise writes as
 *                    ASCII. 'colorstyle' single-handedly determines the
 *                    type of file written:
 *                    if colorstyle==0, (Full Color) a PPM file is written,
 *                    if colorstyle==1, (Greyscale) a PGM file is written,
 *                    if colorstyle==2, (B/W stipple) a PBM file is written.
 */
int
WritePBM ( FILE *fp, byte *pic, const int w, const int h, const int colorstyle,
                     const int bpp, const int raw, char *comment )
{
    int   magic;
    byte *pix;
    int   i, j, len;
    char  buf[BUFSIZ];

    /* Print info to user */
    sprintf (buf , "%4dx%-4d %s, %s format.", w, h,
            (colorstyle==0) ? "PPM" : (colorstyle==1) ? "PGM" : "PBM",
            (raw) ? "raw" : "ascii");
    fprintf (stdout, "\n[%s]", buf);

    /* Calc the appropriate magic number for this file type */
    magic = 0;
    if      (colorstyle==0) magic = 3;
    else if (colorstyle==1) magic = 2;
    else if (colorstyle==2) magic = 1;

    if (raw && magic) magic+=3;


    /* Write the header info */
    fprintf (fp, "P%d\n",magic);
    fprintf (fp, "# Creator: flwt %s\n", REVDATE);

    if (comment) {      /* write comment lines */
        char *sp;

        sp = comment;
        while (*sp) {
            fprintf (fp, "# ");
            while (*sp && *sp != '\n') fputc((int)*sp++, fp);
            if (*sp == '\n') sp++;
            fputc('\n', fp);
        }
    }


    fprintf (fp, "%d %d\n", w, h);
    if ( colorstyle!=2 )
        fprintf (fp, "255\n");

    if ( ferror(fp) )
        return 0;

    /* Write the image data */

    if ( colorstyle==0 ) {                  /* 24bit RGB, 3 bytes per pixel */
        for ( i=0, pix=pic, len=0; i<h; i++ ) {
            for ( j=0 ; j<w ; j++ ) {
	        if ( raw ) {
                    if ( bpp == 24 ) {
	                fputc((int)pix[0], fp);
                        fputc((int)pix[1], fp);
                        fputc((int)pix[2], fp);
                    } else {
                        fputc((int)*pix, fp);
                        fputc((int)*pix, fp);
                        fputc((int)*pix, fp);
                    }
	        } else {
                    if ( bpp == 24 )
	                fprintf (fp, "%3d %3d %3d ", pix[0], pix[1], pix[2]);
                    else
	                fprintf (fp, "%3d %3d %3d ", *pix, *pix, *pix);
	            len += 12;
	            if ( len>58 ) { fprintf (fp, "\n");  len=0; }
	        }
                pix += (bpp==24) ? 3 : 1;
            }
        }
    } else if (colorstyle==1) {             /* 8-bit greyscale */
        for ( i=0, pix=pic, len=0 ; i<w*h ; i++ ) {
            if (raw) {
                fputc ( (bpp==24) ? LUMA(pix[0],pix[1],pix[2]) : *pix, fp);
            } else {
                fprintf (fp, "%3d ", (bpp==24) ? LUMA(pix[0],pix[1],pix[2]) :
                                                 *pix);
	        len += 4;
                if ( len>66 ) { fprintf (fp, "\n");  len=0; }
            }
            pix += (bpp==24) ? 3 : 1;
        }
    } else if (colorstyle==2) {             /* 1-bit B/W stipple */
        int bit, k, flipbw;
        char *str0, *str1;

        /* Shouldn't happen */
        if ( bpp == 24 ) {
            Error ("WritePBM", CUSTOM, RETURN, "24 bpp and B/W Stipple.\n");
            Error ("WritePBM", FATAL_ERROR, ABORT);
            /* NOTREACHED */
        }

        /* if '0' is black, set flipbw */
        flipbw = 1;

        str0 = (flipbw) ? "1 " : "0 ";
        str1 = (flipbw) ? "0 " : "1 ";

        bit = k = 0;   /* Initial values */
        for ( i=0, pix=pic, len=0 ; i<h ; i++ ) {
            for ( j=0, bit=0, k=0 ; j<w ; j++, pix++ ) {
	        if ( raw ) {
	            k = (k << 1) | *pix;
	            bit++;
	            if ( bit==8 ) {
	                if ( flipbw ) k = ~k;
	                fputc(k, fp);
	                bit = k = 0;
	            }
	        } else {
	            if (*pix)
                        fprintf (fp, str1);
	            else
                        fprintf (fp, str0);
	            len += 2;
	            if ( len>68 ) { fprintf (fp, "\n"); len=0; }
	        }
            } /* j */
            if (raw && bit) {
	        k = k << (8-bit);
	        if (flipbw) k = ~k;
	        fputc(k, fp);
            }
        }
    }

    if ( ferror(fp) )
        return 0;

    return 1;
}
